sudo dnf -y install emacs
sudo dnf -y install zlib
sudo dnf -y install zlib-static
sudo dnf -y install openssl-devel
sudo dnf -y install boost-devel
sudo dnf -y install cairo-devel libpcap-devel
sudo dnf -y install libpcap-devel
sudo dnf -y install cairo-devel
