sudo apt-get update --fix-missing
sudo apt-get install -y emacs
sudo apt-get install -y git gcc g++ llvm-3.7 automake autoconf libpcap-dev libboost-dev openssl libssl-dev \
                        zlib1g-dev libcairo2 libcairo2-dev zlibc zlib1g-dev


